from getpass import getpass
import os
import json
import pprint
import time
import pyreadline3

from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.completion import word_completer, WordCompleter
from prompt_toolkit import prompt
from blessed import Terminal


if __name__ != "__main__":
    from ..socketopts import schemas


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(f"{__file__}")))
saved_command = os.path.join(__location__, r'entered_command.json')


LINE_READER = base_read_line = pyreadline3.BaseReadline()


class Formatters:
    """
    Format received dictionaries in the client code
    """
    def print_response(self, input_data, depth: int=0):
        if isinstance(input_data, (list, set, tuple)):
            for data in input_data:
                if isinstance(data, (list, dict, set, tuple)):
                    print("\n")
                self.print_response(data, depth=depth+1)
        elif isinstance(input_data, dict):
            for name, data in input_data.items():
                if isinstance(data, (int, str)):
                    print(("  " * depth) + f"{name}: {data}")
                    continue
                print(("  " * depth) + f"{name}: ")
                self.print_response(data, depth=depth+1)
        elif isinstance(input_data, (int, str)):
            print(f"{depth * '  '}{input_data}")
        if depth == 0:
            print("\n")


class ParserTypeLenEnforcer:
        def __init__(self, name: str=str(), size: tuple=(0, 0), data_type: str='string', choices: list=list()):
            self.arg_name = name
            self.data_type = data_type
            self.lower_size_limit, self.upper_size_limit = size
            self.choices = choices

        def update_constraints(self, name=None, data_type=None, choices=None, size_limit=None, **kwargs):
            self.arg_name = name
            self.data_type = data_type
            self.choices = choices
            self.lower_size_limit, self.upper_size_limit = size_limit

        def __call__(self, data):
            if self.data_type:
                try:
                    if self.data_type == 'string':
                        str(data)
                    elif self.data_type == 'int':
                        int(data)
                except Exception as e:
                    raise ValidationError(message=str(e) + self.data_type, cursor_position=0)
            if self.data_type == int:
                if not data >= self.lower_size_limit or not data < self.upper_size_limit:
                    raise ValidationError(message=f"The input for the argument {self.arg_name} must be in the range ({self.lower_size_limit}, {self.upper_size_limit}). Got value {data}", cursor_position=0)
            elif self.data_type == str:
                if not len(data) >= self.lower_size_limit or not len(data) < self.upper_size_limit:
                    raise ValidationError(message=f"The input length for the argument {self.arg_name} must be in the range ({self.lower_size_limit}, {self.upper_size_limit}). Got length {len(data)}", cursor_position=0)
            if self.choices and data not in self.choices:
                raise ValidationError(message=f"The input for the argument {self.arg_name} must be one of the following: {self.choices}", cursor_position=0)


class ResponseValidator(Validator):
    def __init__(self):
        super().__init__()
        self.enforcer = ParserTypeLenEnforcer()

    def validate(self, document):
        text = document.text
        self.enforcer(text)

        
class Handlers(Formatters):
    def __init__(self):
        self.validator = ResponseValidator()

    def __expression_input(self) -> str: 
        """
        Input an expression as requested by the server for something like cypher queries
        """
        print("Enter 'exit' to submit") 
        expression = ''
        line = ''
        while line != 'exit': 
            line = str(input("> "))
            if line != 'exit':
                expression += line
        return expression
    
    def confirmation_handler(self, argument):
        if 'description' in argument and argument['description']:
            print(argument['description'])
        print(argument['name'])
        while True:
            decision = str(input("(y/n)"))
            if decision == 'y':
                decision = True
                break
            elif decision == 'n':
                decision = False
                break
            else:
                print("Enter valid response")
        return decision
    
    def input_dict_handler(self, term: Terminal, attrs: dict, default=None):
        mode = 'create'
        if not default:
            default = dict()
        answer = default
        default_name = attrs['data_type']['name']
        with term.fullscreen():
            print(f"You are now entering data for {attrs['name']}")
            while True:
                print(f"{term.clear}now editing the map: {attrs['name']}\nreserved names: create, delete, exit (enter these for special action)\nmode: {mode}")
                print(json.dumps(answer, indent=4))
                attrs['data_type']['name'] = default_name
                name = str(input(f"enter the name for the instance of your {attrs['data_type']['name']}: "))
                if name.lower() in ('delete', 'create'):
                    mode = name
                    continue
                if name.lower() == 'exit':
                    return answer
                if mode == 'create':
                    mode = 'create'
                    attrs['data_type']['name'] = name
                    if name in answer:
                        default_value = answer#[name]
                    else:
                        default_value = None
                    pprint.pprint({name:attrs['data_type']})
                    sub_answer = self.advanced_input_handler({name:attrs['data_type']}, term, default=default_value)
                    answer.update(**sub_answer)
                elif mode == 'delete':
                    mode = 'delete'
                    try:
                        answer.pop(name)
                    except KeyError:
                        pass

    def input_list(self, term: Terminal, attrs: dict, default=None):
        if not default:
            default = list()
        answer = default
        with term.fullscreen():
            print(f"You are now entering data for {attrs['name']}")
            mode = 'modify'
            while True:
                sub_answer = None
                presentable_dict = {str(index+1):value for index, value in enumerate(answer)}
                print(rf"""{term.clear}now editing the list: {attrs['name']}
                      Enter exit to submit list, or new to create new list element.
                      Modes: to change mode, enter the mode you want to use. 
                        Delete: when you enter an index the index is deleted. 
                        Modify: when you enter an index you will re-enter the element to modify it
                      mode: {mode}""")
                print(json.dumps(presentable_dict, indent=4))
                decision = input("Enter 'new' or 'exit': ")
                if decision.lower() == 'exit':
                    return answer
                elif decision.lower() == 'new':
                    pass
                elif decision.lower() in ('modify', 'delete'):
                    mode = decision
                    continue
                elif decision.isdigit() and mode == 'delete':
                    try:
                        answer.pop(int(decision)-1)
                    except IndexError:
                        continue
                    continue
                elif decision.isdigit() and mode == 'modify':
                    try:
                        print({f"{attrs['name']}_{str(decision)}":answer[int(decision)-1]})
                        time.sleep(5)
                        sub_answer = self.advanced_input_handler({f"{attrs['name']}_{str(decision)}":attrs['data_type']}, term, default={f"{attrs['name']}_{str(decision)}":answer[int(decision)-1]})
                        answer[int(decision)-1] = sub_answer[f"{attrs['name']}_{str(decision)}"]
                    except IndexError:
                        continue
                    continue
                else:
                    continue
                sub_answer = self.advanced_input_handler({f"{attrs['name']}_{str(len(answer)+1)}":attrs['data_type']}, term)
                index, value = list(sub_answer.items())[0]
                answer.append(value)

    def form_handler(self, term: Terminal, attrs: dict, form_name, default=None):
        form_options = attrs['arguments_list']
        completer = WordCompleter(list(form_options.keys()))
        if not default:
            form_input = dict()
            for name, command_metadata in form_options.items():
                if command_metadata['default_value']:
                    form_input[name] = command_metadata['default_value']
                    continue
                form_input[name] = ''
        else:
            form_input = {arg_name:arg_default_value for arg_name, arg_default_value in default.items() if arg_name in form_options}
        while True:
            print(f"{term.clear}now editing the form: {form_name}")
            print(json.dumps(form_input, indent=3))
            field = prompt('Enter the field you want to modify. Enter exit to complete: ', completer=completer)
            if field.lower() == 'exit':
                return form_input
            elif field not in form_options:
                continue
            result = self.advanced_input_handler({field:form_options[field]}, term, default=form_input)
            form_input[field] = result[field]
            
    def str_input(self, attrs):
        completer = None
        if 'choices' in attrs and attrs['choices']:
            completer = WordCompleter(attrs['choices'])
        answer = prompt(f"{attrs['name']}: ", validator=self.validator, wrap_lines=True, completer=completer)
        return answer
    
    def advanced_input_handler(self, form_request: dict, term: Terminal, default=None):
        response = dict()
        for field, attrs in form_request.items():
            arg_type = attrs['arg_type']
            self.validator.enforcer.update_constraints(**attrs)
            if 'description' in attrs and attrs['description']:
                print(attrs['description'])
            # pprint.pprint(form_request)
            # print(field)
            if default:
                default_selection = default[field]
            else:
                default_selection = None
            try:
                match arg_type:
                    case 'secure':
                        answer = prompt(f"{attrs['name']}: ", validator=self.validator, is_password=True)
                    case 'expression':
                        with term.fullscreen():
                            print(f"Enter expression input for the {attrs['name']} argument.")
                            answer = self.__expression_input()
                    case 'form':
                        answer = self.form_handler(term, attrs, field, default=default_selection)
                    case 'input_list':
                        answer = self.input_list(term, attrs, default=default_selection)
                    case 'input_dict':
                        answer = self.input_dict_handler(term, attrs, default=default_selection)
                    case 'str_input':
                        answer = self.str_input(attrs)
                    case 'standard':
                        answer = self.str_input(attrs)
                    case 'confirmation':
                        answer = self.confirmation_handler(attrs)
                    case 'silent':
                        continue
                    case _:
                        raise AttributeError(f"There is not argument type {arg_type}")
                response[field] = answer
            except KeyboardInterrupt:
                raise RuntimeError('Form cancelled, command execution stopped')
            except Exception as e:
                with open(saved_command, 'a') as f:
                    f.write(json.dumps(response))
                    print(f"Argument input failure, command data written to file {saved_command}")
                    raise e
        return response
    
    def universal_message_handler(self, message: schemas.BaseSchema, term: Terminal):
        self.print_response(message.message)
        if message.error:
            self.print_response(message.error)
        if message.request_content:
            try:
                default = message.existing_data
            except:
                default = None
            filled_form = self.advanced_input_handler(message.request_content, term, default=default)
            return filled_form
        return None
        
    def environment_cli_response_stream_handler(self, response):
        if response.schema_type == 'ResponseData' and response.exit_status: # if the command was a shutdown or exit, close the program
            self.print_response(response.response_message)
            os._exit(0)
        elif response.schema_type == 'ResponseData':
            if response.active_username:
                self.username = response.active_username
            if response.url:
                self.url = response.url
            self.print_response(response.response_message)


import typing
import abc


ALLOWED_ARG_TYPES = typing.Literal['silent', 'secure', 'expression', 'input_list', 'input_dict', 'form', 'str_input', 'confirmation']
ALLOWED_DATA_TYPES = typing.Literal['string', 'int', 'bool']
ALLOWED_ACTIONS = typing.Literal['store', 'store_true', 'store_false']


class DynamicChoiceList(abc.ABC):
    @abc.abstractmethod
    def __call__(self, tapis_instance):
        pass


class AbstractArgument(abc.ABC):
    @abc.abstractmethod
    def json(self):
        pass


def cast_int(data):
    return int(data)

def cast_string(data):
    return str(data)

def cast_bool(data):
    return bool(data)


class Argument(AbstractArgument):
    type_map = {
        'int':cast_int,
        'string':cast_string,
        'bool':cast_bool
    }
    def __init__(self, argument: str,
                 data_type: ALLOWED_DATA_TYPES | typing.Type[AbstractArgument] = 'string',
                 arg_type: ALLOWED_ARG_TYPES | typing.Literal['standard']='standard',
                 choices: list | typing.Type[DynamicChoiceList] | None=None, 
                 action: typing.Literal['store', 'store_true', 'store_false']="store", 
                 description: str="",
                 positional: bool=False,
                 default_value=None,
                 depends_on: list = [],
                 size_limit: tuple=(0,4096)):
        if arg_type not in typing.get_args(ALLOWED_ARG_TYPES) and arg_type != 'standard':
            raise ValueError(f"The arg type parameter in the argument {self.__class__.__name__} must be in the list {ALLOWED_ARG_TYPES}. Got arg type {arg_type}")
        if data_type not in typing.get_args(ALLOWED_DATA_TYPES) and not issubclass(data_type.__class__, AbstractArgument):
            raise ValueError(f"The data type argument provided to the argument {self.__class__.__name__} must be in the list {ALLOWED_DATA_TYPES}, or must be a subclass of AbstractArgument")
        if action not in typing.get_args(ALLOWED_ACTIONS):
            raise ValueError(f"Action {action} not in the list {ALLOWED_ACTIONS}, in argument {self.__class__.__name__}")
        if arg_type in ('input_list', 'input_dict') and isinstance(data_type, Argument) and data_type.arg_type == 'standard':
            data_type.arg_type = 'str_input'
        if action != 'store':
            data_type = 'bool'
        self.argument = argument
        self.arg_type = arg_type
        self.data_type = data_type
        self.choices = choices
        if self.arg_type != 'standard':
            action = 'store_true'
        self.action = action
        self.description = description
        self.positional = positional
        self.default_value = default_value
        self.size_limit=size_limit
        self.depends_on = depends_on

        self.truncated_arg = None
        self.full_arg = f"--{self.argument}"

    def verify_standard_value(self, value):
        if self.arg_type == "standard":
            #print(f"ARGUMENT NAME: {self.argument}\nARGUMENT VALUE: {value}\n")
            if value == None and self.default_value:
                value = self.default_value
                return value
            elif not value:
                return value
            min_, max_ = self.size_limit
            try:
                value = self.type_map[self.data_type](value)
            except:
                raise ValueError(f"The argument {self.argument} requires a datatype {self.data_type}")
            if type(value) == int:
                if value >= max_ or value < min_:
                    raise ValueError(f"The argument {self.argument} must be a value in the range {self.size_limit}")
            elif type(value) == str and (len(value) >= max_ or len(value) < min_):
                raise ValueError(f"The argument {self.argument} must be between the sizes {self.size_limit}")
            elif self.choices and value not in self.choices:
                raise ValueError(f"The value for argument {self.argument} must be in the list {self.choices}")
            elif value == None and self.default_value:
                value = self.default_value
        return value

    def json(self):
        json = {
            'name':self.argument,
            'arg_type':self.arg_type,
            'data_type':self.data_type,
            'choices':self.choices,
            'action':self.action,
            'description':self.description,
            'positional':self.positional,
            'default_value':self.default_value,
            'size_limit':self.size_limit,
            'truncated_arg':self.truncated_arg,
            'full_arg':self.full_arg,
            'depends_on':self.depends_on
        }
        if self.data_type in ('string', 'int', 'bool'):
            json['data_type'] = self.data_type
        else:
            json['data_type'] = self.data_type.json()
        return json
    
    def help_message(self):
        help = {"name":self.argument,
                "description":f"{self.description}"}
        if self.choices and isinstance(self.choices, list):
            help['choices'] = self.choices
        if self.positional:
            help['syntax'] = f"<{self.argument}>"
        elif self.action == 'store':
            help['syntax'] = f"{self.truncated_arg}/{self.full_arg} <{self.argument}>"
        else:
            help['syntax'] = f"{self.truncated_arg}/{self.full_arg}"
            if self.arg_type != 'standard':
                help['syntax'] = help['syntax'] + " (f)"
        return help
    
    def check_for_copy_data(self):
        return {
            'name':self.argument,
            'action':self.action,
        }
    
    def str(self):
        help_str = f"{self.truncated_arg}/{self.full_arg} "
        if self.positional:
            help_str = f"<{self.argument}> "
        elif self.action == 'store':
            help_str += f"<{self.argument}> "
        if self.arg_type != 'standard':
            help_str += ' (f)\t'
        return help_str


class Form(Argument):
    def __init__(self, form_name, arguments_list, description=None):
        super().__init__(form_name, arg_type='form', description=description)
        for argument in arguments_list:
            if argument.arg_type == 'standard':
                argument.arg_type = 'str_input'
        self.arguments_list = {argument.argument:argument for argument in arguments_list}

    def json(self):
        fields = {argument_name:argument.json() for argument_name, argument in self.arguments_list.items()}
        json = {
            'name':self.argument,
            'arg_type':self.arg_type,
            'data_type':self.data_type,
            'choices':self.choices,
            'action':self.action,
            'description':self.description,
            'positional':self.positional,
            'default_value':self.default_value,
            'size_limit':self.size_limit,
            'truncated_arg':self.truncated_arg,
            'full_arg':self.full_arg,
            'arguments_list':fields
        }
        return json
    

class RequestHandler:
    def handle_requests(self, request):
        pass
    

if __name__ == "__main__":
    pass
    




import json


if __name__ != "__main__":
    from . import baseCommand, decorators
    from .arguments import argument
    Argument = argument.Argument


def get_latest_version(t, kwargs):
    if 'appVersion' not in kwargs:
        version = t.apps.getAppLatestVersion(**kwargs).version
        kwargs['appVersion'] = version
    return kwargs


class create_app(baseCommand.BaseCommand):
    """
    @help: create an app. You must have a properly configured app config file. 
    See a template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/app-config.json
    """
    supports_config_file=True
    required_arguments = [
        Argument('id', positional=True, description='This can be the same as another app as long as the version number is different'),
        Argument('version', positional=True),
        Argument('containerImage')
    ]
    optional_arguments = [
        Argument('description', arg_type='str_input'),
        Argument('owner', default_value=r"${apiUserId}"),
        Argument('enabled', action='store_true'),
        Argument('runtime', choices=['SINGULARITY', 'DOCKER']),
        Argument('runtimeVersion'),
        Argument('runtimeOptions', choices=['NONE', 'SINGULARITY_START', 'SINGULARITY_RUN']),
        Argument('jobType', choices=['BATCH', 'FORK']),
        Argument('maxJobs', data_type='int'),
        Argument('maxJobsPerUser', data_type='int'),
        Argument("strictFileInputs", action='store_true'),
        Argument('tags', arg_type='input_list', data_type=Argument('tag', size_limit=(1, 128))),
    ]
    async def run(self, *args, **kwargs) -> str: # create a tapis app taking a json descriptor file path
        url = self.t.apps.createAppVersion(**kwargs)
        return f"App created successfully\nID: {kwargs['id']}\nVersion: {kwargs['version']}\nURL: {url}\n"
    

class AppUpdatingRetriever(baseCommand.UpdatableFormRetriever):
    def __call__(self, tapis_instance, **kwargs):
        kwargs = get_latest_version(tapis_instance, kwargs)
        app_data = tapis_instance.apps.getApp(appId=kwargs['appId'], appVersion=kwargs['appVersion'])
        return app_data
    

class update_app(create_app):
    """
    @help: update app with the select attributes
    """
    updateable_form_retriever = AppUpdatingRetriever()
    supports_config_file = True
    required_arguments = [
        Argument('appId', size_limit=(1, 80), positional=True),
    ]
    optional_arguments = [
        Argument('appVersion', size_limit=(1, 64)),
        Argument('description', arg_type='str_input'),
        Argument('owner', default_value=r"${apiUserId}"),
        Argument('enabled', action='store_true'),
        Argument('runtime', choices=['SINGULARITY', 'DOCKER']),
        Argument('runtimeVersion'),
        Argument('runtimeOptions', choices=['NONE', 'SINGULARITY_START', 'SINGULARITY_RUN']),
        Argument('jobType', choices=['BATCH', 'FORK']),
        Argument('maxJobs', data_type='int'),
        Argument('maxJobsPerUser', data_type='int'),
        Argument("strictFileInputs", action='store_true'),
        Argument('tags', arg_type='input_list', data_type=Argument('tag', size_limit=(1, 128))),
    ]
    async def run(self, *args, **kwargs):
        kwargs = get_latest_version(self.t, kwargs)
        self.t.apps.patchApp(**kwargs)
        return f'updated app {kwargs["appId"]} successfully'
    

class JobUpdatingRetriever(baseCommand.UpdatableFormRetriever):
    def __call__(self, tapis_instance, **kwargs):
        kwargs = get_latest_version(tapis_instance, kwargs)
        app_data = tapis_instance.apps.getApp(appId=kwargs['appId'], appVersion=kwargs['appVersion']).jobAttributes
        return app_data


class assign_default_job_attributes(baseCommand.BaseCommand):
    """
    @help: assign a default set of job attributes for this app, if you decide to run it as a job
    """
    updateable_form_retriever = JobUpdatingRetriever()
    supports_config_file=True
    required_arguments = [
        Argument('appId', size_limit=(1, 80), positional=True)
    ]
    optional_arguments = [
        Argument('appVersion', size_limit=(1, 64)),
        Argument('description', arg_type='str_input'),
        Argument('dynamicExecSystem', arg_type='confirmation', description="System is dynamic?"),
        Argument('execSystemConstraints', arg_type='input_list', data_type=Argument('constraint', size_limit=(3, 4096))),
        Argument('execSystemId', size_limit=(1, 80), description='what system id will this be run on?'),
        Argument('execSystemExecDir', size_limit=(1, 4096), description='based on the system job working dir? automatically generated if not specified'),
        Argument('execSystemInputDir', size_limit=(1, 4096), description='what system path will be used to stage input files? automatically generated if not specified'),
        Argument('execSystemOutputDir', size_limit=(1, 4096), description='Where will tapis put job output files? automatically generated if not specified'),
        Argument('execSystemLogicalQueue', size_limit=(1, 128), description='What batch logical queue on the system will be used for execution?'),
        Argument('archiveSystemId', size_limit=(1, 80), description='What system will be used when archiving outputs?'),
        Argument('archivesystemDir', size_limit=(1, 4096), description='What system directory will be used for archiving?'),
        Argument('archiveOnAppError', arg_type='confirmation', description='archive on error?'),
        Argument("isMpi", arg_type='confirmation', description="is mpi?"),
        Argument('mpiCmd', size_limit=(1, 126), arg_type='str_input', description='command to launch MPI jobs. Conflicts with cmdPrefix if isMpi is set'),
        Argument('cmdPrefix', size_limit=(1, 126), description='string put in front of app run command to run with mpi command'),
        argument.Form('parameterSet', arguments_list=[
            Argument('appArgs', arg_type='input_list', data_type=argument.Form('appArg', arguments_list=[
                Argument('name', size_limit=(1, 80)),
                Argument('description', size_limit=(1, 8096)),
                Argument('inputMode', choices=['REQUIRED', "FIXED", "INCLUDE_ON_DEMAND", "INCLUDE_BY_DEFAULT"]),
                Argument('arg')]
            ), description='command line arguments to be passed to the application'),
            Argument('containerArgs', arg_type='input_list', data_type=argument.Form('containerArg', arguments_list=[
                Argument('name', size_limit=(1, 80)),
                Argument('description', size_limit=(1, 8096)),
                Argument('inputMode', choices=['REQUIRED', "FIXED", "INCLUDE_ON_DEMAND", "INCLUDE_BY_DEFAULT"], description='How will this argument be treated when processing jobs? EG should the job require it?'),
                Argument('arg', description='value for the argument')
            ]), description='command line arguments to be passed to the container at runtime, whether singularity or container'),
            Argument('schedulerOptions', arg_type='input_list', data_type=argument.Form('schedulerOption', arguments_list=[
                Argument('name', size_limit=(1, 80)),
                Argument('description', size_limit=(1, 8096)),
                Argument('inputMode', choices=['REQUIRED', "FIXED", "INCLUDE_ON_DEMAND", "INCLUDE_BY_DEFAULT"]),
                Argument('arg')
            ]), description='options to pass to hpc batch scheduler'),
            Argument('envVariables', arg_type='input_list', data_type=argument.Form('environment_variable', arguments_list=[
                Argument('key'),
                Argument('value'),
                Argument('description', size_limit=(1, 2048))
            ]), description='environment variables placed into the runtime environment')
        ], description='collections used during job execution. Specify app args, container args, scheduler options, evnironment variables, and archive filter for job execution'),
        Argument('fileInputs', arg_type='input_list', data_type=argument.Form('fileInput', arguments_list=[
            Argument('name', size_limit=(1, 80)),
            Argument('description', size_limit=(1, 8096)),
            Argument('inputMode', choices=['REQUIRED', 'OPTIONAL', 'FIXED'], description='indicates how input should be treeated when processing job request'),
            Argument('autoMountLocal', arg_type='confirmation', description='should the service automatically mount file paths to the containers?'),
            Argument('sourceUrl', description='source used by jobs service when staging file inputs'),
            Argument('targetPath', description='where in the container will the files go when staging the inputs')
        ]), description='selection of file inputs that must be staged to run application'),
        Argument('fileInputArrays', arg_type='input_list', data_type=argument.Form('fileInput', arguments_list=[
            Argument('name', size_limit=(1, 80)),
            Argument('description', size_limit=(1, 8096)),
            Argument('inputMode', choices=['REQUIRED', 'OPTIONAL', 'FIXED']),
            Argument('sourceUrls', arg_type='input_list', data_type=Argument('sourceUrl')),
            Argument('targetPath')
        ]), description='collection of arrays of inputs to be staged to the application'),
        Argument('nodeCount', data_type='int', description='how many nodes do you want to request in your job?'),
        Argument('coresPerNode', data_type='int', description='how many cores per node?'),
        Argument('memoryMB', data_type='int'),
        Argument('maxMinutes', data_type='int', description='max job runtime'),
        Argument('subscriptions', arg_type='input_list', data_type=argument.Form('subscription', arguments_list=[
            Argument('description'),
            Argument('enabled', arg_type='confirmation', description='enable subscription?'),
            Argument('jobEventCategoryFilter', choices=['ALL', 'JOB_NEW_STATUS', 'JOB_INPUT_TRANSACTION_ID', 'JOB_ARCHIVE_TRANSACTION_ID', 'JOB_ERROR_MESSAGE', 'JOB_SUBSCRIPTION']),
            Argument('ttlMinutes', data_type='int'),
            Argument('deliveryTargets', arg_type='input_list', data_type=argument.Form('deliveryTarget', arguments_list=[
                Argument('deliveryMethod', choices=['WEBHOOK', 'EMAIL']),
                Argument('deliveryAddress')
            ]))
        ]))
    ]
    async def run(self, *args, **kwargs):
        kwargs = get_latest_version(self.t, kwargs)
        appId = kwargs.pop('appId')
        appVersion = kwargs.pop('appVersion')
        kwargs.pop('connection')
        jobAttributes = {'jobAttributes':kwargs}
        self.t.apps.patchApp(appId=appId, appVersion=appVersion, **jobAttributes)
        return 'updated job attributes successfully'
    

class get_apps(baseCommand.BaseCommand):
    """
    @help: get a list of all available apps
    """
    return_fields = ['id', 'version']
    optional_arguments = [
        Argument('listType', choices=[
            'OWNED',
            'SHARED_PUBLIC',
            'SHARED_DIRECT',
            'READ_PERM', 
            'MINE',
            'ALL'
        ])
    ]
    async def run(self, *args, **kwargs):
        return self.t.apps.getApps(**kwargs)
    

class get_app(baseCommand.BaseCommand):
    """
    @help: get a specific app
    """
    return_fields = ['id', 'version', 'containerImage']
    required_arguments = [
        Argument('appId', size_limit=(1, 80), positional=True),
    ]
    optional_arguments = [
        Argument('appVersion', size_limit=(1, 64))
    ]
    async def run(self, *args, **kwargs):
        kwargs = get_latest_version(self.t, kwargs)
        return self.t.apps.getApp(**kwargs)
    

class is_app_enabled(baseCommand.BaseCommand):
    """
    @help: check if the app is enabled
    """
    required_arguments = [
        Argument('appId', size_limit=(1, 80), positional=True),
    ]
    async def run(self, *args, **kwargs):
        return self.t.apps.isEnabled(**kwargs)
    

class enable_app(is_app_enabled):
    """
    @help: enable the app
    """
    async def run(self, *args, **kwargs):
        return self.t.apps.enableApp(**kwargs)
    

class disable_app(is_app_enabled):
    """
    @help: disable the app
    """
    async def run(self, *args, **kwargs):
        return self.t.apps.disableApp(**kwargs)
    

class delete_app(is_app_enabled):
    """
    @help: delete the app
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, *args, **kwargs):
        return self.t.apps.disableApp(**kwargs)
    

class undelete_app(is_app_enabled):
    """
    @help: undo a deletion of an app
    """
    async def run(self, *args, **kwargs):
        return self.t.apps.undeleteApp(**kwargs)
    

class get_app_history(is_app_enabled):
    """
    @help: get the history of changes to an application
    """
    async def run(self, *args, **kwargs):
        return str(self.t.apps.getAppHistory(**kwargs))
    

class get_app_user_perms(baseCommand.BaseCommand):
    """
    @help: get list of user permissions for the app
    """
    required_arguments = [
        Argument('appId', size_limit=(1, 80), positional=True),
        Argument('userName', size_limit=(1, 80))
    ]
    async def run(self, *args, **kwargs):
        return str(self.t.apps.getUserPerms)
    

class grant_app_user_perms(baseCommand.BaseCommand):
    """
    @help: grant user perms for a specific user on an app
    """
    required_arguments = [
        Argument('appId', size_limit=(1, 80), positional=True),
        Argument('userName', size_limit=(1, 80)),
        Argument('permissions', arg_type='input_list', data_type=Argument('permission', choices=['READ', 'MODIFY', 'EXECUTE']))
    ]
    async def run(self, *args, **kwargs):
        return self.t.apps.grantUserPerms(**kwargs)
    

class revoke_app_user_perms(grant_app_user_perms):
    """
    @help: revoke perms for a specific user on an app
    """
    async def run(self, *args, **kwargs):
        return self.t.apps.revokeUserPerms(**kwargs)


import pyperclip
from tapipy.tapis import errors as TapisErrors


if __name__ != "__main__":
    from . import baseCommand, decorators
    from .arguments import argument
    Argument = argument.Argument


class get_pods(baseCommand.BaseCommand):
    """
    @help: return a list of pods the current tapis instance has access to
    """
    return_fields = ['pod_id', 'pod_template', 'status']
    async def run(self, *args, **kwargs) -> str: 
        pods_list = self.t.pods.get_pods()
        return pods_list


class get_pod(baseCommand.BaseCommand):
    """
    @help: return a specific pod based on pod_id
    """
    return_fields = ['pod_id', 'pod_template', 'status']
    required_arguments=[
        Argument('pod_id', positional=True)
    ]
    async def run(self, *args, **kwargs) -> str: 
        pod_data = self.t.pods.get_pod(pod_id=kwargs["pod_id"])
        return pod_data


class create_pod(baseCommand.BaseCommand):
    """
    @help: create a new pod on the selected Tapis service
    @doc: fix the pod updating, make sure non selected optional vaiables do not overwrite. Why description appending???
    """
    return_fields = ['pod_id', 'pod_template', 'status']
    supports_config_file=True
    required_arguments=[
        Argument('pod_id', positional=True),
        Argument('pod_template', positional=True)
    ]
    optional_arguments=[
        Argument('description', arg_type='str_input', size_limit=(0, 2048)),
        Argument('command', arg_type='input_list', data_type=argument.Argument('command', arg_type='str_input'), description="Command to run in the pod"),
        Argument('environment_variables', arg_type='input_dict', data_type=argument.Argument('environment_variable', arg_type='str_input')),
        Argument('roles_required', arg_type='input_list', data_type=argument.Argument('required_role', arg_type='str_input'), description='what role is required by the user to access this pod?'),
        Argument('time_to_stop_default', data_type='int'),
        Argument('time_to_stop_instance', data_type='int'),
        Argument('volume_mounts', arg_type='input_dict', data_type=argument.Form(
            'volume_mount', arguments_list = [
                Argument('type', choices=['tapisvolume', 'tapissnapshot', 'pvc']),
                Argument("mount_path", description='This is top level path you want to mount the volume on inside the pod. This is something like <neo4j-home>\data for a neo4j pod. Data from that path will load to the mount and become persistent'),
                Argument('sub_path', description='If you want to only load a single file, like file.txt (which is inside the parent mount path) you can specify here')
            ]
        ), description="Used to attach the pod to an existing kubernetes volume to provide pod persistence (in case of crash). Each key is the volume_id"),
        Argument('networking', arg_type='input_dict', data_type=argument.Form(
            'network', arguments_list = [
                Argument('protocol', description='Something like https'),
                Argument('port', data_type='int'),
                Argument('url')
            ]
        ), description='Important networking configuration. You probably shouldnt touch this, but I wont stop you'),
        argument.Form(
            'resources', arguments_list = [
                Argument('cpu_request', data_type='int'),
                Argument('cpu_limit', data_type='int'),
                Argument('mem_request', data_type='int'),
                Argument('mem_limit', data_type='int'),
            ]
        )
    ]
    async def run(self, *args, **kwargs) -> str:
        template_name = kwargs['pod_template']
        template_formats = (f"template/{template_name}", f"{self.username}/{template_name}", template_name)
        if 'volume_mounts' in kwargs and kwargs['volume_mounts']:
            for mount_name, mount in kwargs['volume_mounts'].items():
                if mount['type'] == 'tapisvolume':
                    try:
                        self.t.pods.get_volume(volume_id=mount_name)
                    except:
                        self.t.pods.create_volume(volume_id=mount_name)
                elif mount['type'] == 'tapissnapshot':
                    try:
                        self.t.pods.get_snapshot(snapshot_id=mount_name)
                    except:
                        raise Exception('snapshot not found')
        for template_format in template_formats:
            try:
                kwargs['pod_template'] = template_format
                results = self.t.pods.create_pod(**kwargs)
                break
            except TapisErrors.BadRequestError as e:
                if template_format != template_formats[-1]:
                    continue
                raise ValueError(f"Failed to execute pod creation due to {str(e)}")
        return results
    

class PodUpdatingRetriever(baseCommand.UpdatableFormRetriever):
    def __call__(self, tapis_instance, **kwargs):
        pod_data = tapis_instance.pods.get_pod(pod_id=kwargs["pod_id"])
        return pod_data
    

class update_pod(create_pod): # make it so the command retrieves current settings and sends them over so forms can append and edit instead of just overwrite
    """
    @help: update a pod. Must be restarted to stage changes
    """
    updateable_form_retriever=PodUpdatingRetriever()
    return_fields = ['pod_id', 'pod_template', 'status']
    required_arguments=[
        Argument('pod_id', positional=True),
    ]
    async def run(self, *args, **kwargs):
        print(kwargs)
        pod_information = self.t.pods.update_pod(**kwargs)
        return pod_information


class start_pod(baseCommand.BaseCommand):
    """
    @help: start the pod specified with pod_id
    """
    return_fields = ['pod_id', 'pod_template', 'status']
    required_arguments=[
        Argument('pod_id', positional=True),
    ]
    async def run(self, *args, **kwargs):
        return_information = self.t.pods.start_pod(pod_id=kwargs['pod_id'])
        return return_information


class restart_pod(baseCommand.BaseCommand):
    """
    @help: initiate a pod restart
    """
    return_fields = ['pod_id', 'pod_template', 'status']
    decorator=decorators.NeedsConfirmation()
    required_arguments=[
        Argument('pod_id', positional=True),
    ]
    async def run(self, *args, **kwargs) -> str:
        return_information = self.t.pods.restart_pod(pod_id=kwargs['pod_id'])
        return return_information


class stop_pod(baseCommand.BaseCommand):
    """
    @help: stop a pod's operations
    """
    return_fields = ['pod_id', 'pod_template', 'status']
    decorator=decorators.NeedsConfirmation()
    required_arguments=[
        Argument('pod_id', positional=True),
    ]
    async def run(self, *args, **kwargs):
        return_information = self.t.pods.stop_pod(pod_id=kwargs['pod_id'])
        return str(return_information)


class delete_pod(baseCommand.BaseCommand):
    """
    @help: delete select pod
    """
    return_fields = ['pod_id', 'pod_template', 'status']
    decorator=decorators.NeedsConfirmation()
    required_arguments=[
        Argument('pod_id', positional=True),
    ]
    async def run(self, *args, **kwargs) -> str: 
        return_information = self.t.pods.delete_pod(pod_id=kwargs['pod_id'])
        return str(return_information)
    

class set_pod_perms(baseCommand.BaseCommand):
    """
    @help: set the permissions for the pod selected
    """
    return_fields = ['permissions']
    required_arguments=[
        Argument('pod_id', positional=True),
        Argument('username'),
        Argument('level')
    ]
    async def run(self, *args, **kwargs) -> str: # set pod permissions, given a pod pod_id, user, and permission level
        return_information = self.t.pods.set_pod_permission(pod_id=kwargs['pod_id'], user=kwargs['username'], level=kwargs['level'])
        return return_information


class delete_pod_perms(baseCommand.BaseCommand):
    """
    @help: delete the selected pod from the pods service you are connected to
    """
    return_fields = ['permissions']
    decorator=decorators.NeedsConfirmation()
    required_arguments=[
        Argument('pod_id', positional=True),
        Argument('username')
    ]
    async def run(self, *args, **kwargs) -> str: # take away someones perms if they are being malicious, or something
        return_information = self.t.pods.delete_pod_perms(pod_id=kwargs['pod_id'], user=kwargs['username'])
        return return_information


class get_pod_perms(baseCommand.BaseCommand):
    """
    @help: get the permissions list for the selected pod
    """
    return_fields = ['permissions']
    required_arguments=[
        Argument('pod_id', positional=True),
    ]
    async def run(self, *args, **kwargs) -> str: # return a list of permissions on a given pod
        return_information = self.t.pods.get_pod_permissions(pod_id=kwargs['pod_id'])
        return return_information


class copy_pod_password(baseCommand.BaseCommand):
    """
    @help: copy the pod password to the clipboard
    """
    decorator=decorators.Auth()
    required_arguments=[
        Argument('pod_id', positional=True),
    ]
    async def run(self, *args, **kwargs) -> str: # copies the pod password to clipboard so that the user can access the pod via the neo4j desktop app. Maybe a security risk? not as bad as printing passwords out!
        password = self.t.pods.get_pod_credentials(pod_id=kwargs['pod_id']).user_password
        pyperclip.copy(password)
        password = None
        return 'copied to clipboard'


class get_pod_logs(baseCommand.BaseCommand):
    """
    @help: retrieve the logs of an active pod and either print them to the console, or write them to the specified file
    """
    return_fields = ['logs']
    required_arguments=[
        Argument('pod_id', positional=True)
    ]
    optional_arguments=[
        Argument('destination_file')
    ]
    async def run(self, *args, **kwargs):
        logs = self.t.pods.get_pod_logs(pod_id=kwargs['pod_id'])
        if 'destination_file' in kwargs:
            file = kwargs['destination_file']
        else:
            file = None
        if file:
            with open(file, 'w') as f:
                f.write(logs)
            return f"Log saved at {file}"
        return logs
    




if __name__ != "__main__":
    from . import systemCommands, volumeCommands, serverCommands, appCommands, podCommands, fileCommands, dataFormatters, baseCommand, jobCommands
    from .query import postgres, neo4j
    from utilities import exceptions
    from commands.arguments.argument import Argument
    from commands.commandOpts import CHECK_EXPLICIT_ID


class Systems(baseCommand.BaseCommandMap):
    """
    @help: run operations on Tapis systems
    """
    command_map = {
        'get_systems':systemCommands.get_systems(), # since initialization of commands is separate from __init__, you dont need to specify these as classes anymore
        'get_system_info':systemCommands.get_system_info(),
        'get_scheduler_profiles':systemCommands.get_scheduler_profiles(),
        'submit_system_credentials':systemCommands.submit_system_credentials(),
        'verify_pki_keys':systemCommands.verify_pki_keys(),
        'create_system':systemCommands.create_system(),
        'update_system':systemCommands.update_system(),
        'is_system_enabled':systemCommands.is_system_enabled(),
        'enable_system':systemCommands.enable_system(),
        'disable_system':systemCommands.disable_system(),
        'system':systemCommands.system(),
        'exit_system':systemCommands.exit_system(),
        'delete_system':systemCommands.delete_system(),
        'undelete_system':systemCommands.undelete_system(),
        'create_child_system':systemCommands.create_child_system(),
        'unlink_child_system':systemCommands.unlink_child_system(),
        'unlink_children':systemCommands.unlink_children(),
        'get_user_perms':systemCommands.get_user_perms(),
        'grant_user_perms':systemCommands.grant_user_perms(),
        'revoke_user_perms':systemCommands.revoke_user_perms(),
    }


class General(baseCommand.BaseCommandMap):
    """
    @help: run config operations on the 
    """
    command_map = {
        'get_tenants':serverCommands.get_tenants(),
        'get_tenant':serverCommands.get_tenant(),
        'whoami':serverCommands.whoami(),
        'user':serverCommands.user(),
        'whereami':serverCommands.whereami(),
        'exit':serverCommands.exit(),
        'shutdown':serverCommands.shutdown(),
        "get_args":serverCommands.get_args(),
        'switch_service_to':serverCommands.switch_service_to(),
        'manpages':serverCommands.manpages()
    }


class Pods(baseCommand.BaseCommandMap):
    """
    @help: run operations on tapis pods
    """
    command_map = {
        'get_pods':podCommands.get_pods(),
        'get_pod':podCommands.get_pod(),
        'create_pod':podCommands.create_pod(),
        'update_pod':podCommands.update_pod(),
        'start_pod':podCommands.start_pod(),
        'restart_pod':podCommands.restart_pod(),
        'delete_pod':podCommands.delete_pod(),
        'set_pod_perms':podCommands.set_pod_perms(),
        'stop_pod':podCommands.stop_pod(),
        'delete_pod_perms':podCommands.delete_pod_perms(),
        'get_pod_perms':podCommands.get_pod_perms(),
        'copy_pod_password':podCommands.copy_pod_password(),
        'get_pod_logs':podCommands.get_pod_logs(),
    }


class Volumes(baseCommand.BaseCommandMap):
    """
    @help: run operations on tapis volumes and snapshots
    """
    command_map = {
        'get_volumes':volumeCommands.get_volumes(),
        'create_volume':volumeCommands.create_volume(),
        'get_volume':volumeCommands.get_volume(),
        'volume':volumeCommands.volume(),
        'exit_volume':volumeCommands.exit_volume(),
        'update_volume':volumeCommands.update_volume(),
        'delete_volume':volumeCommands.delete_volume(),
        'dir':volumeCommands.dir(),
        'upload_volume':volumeCommands.upload_volume(),
        'set_volume_permission':volumeCommands.set_volume_permission(),
        'get_volume_permissions':volumeCommands.get_volume_permissions(),
        'delete_volume_permission':volumeCommands.delete_volume_permission(),
        'get_snapshots':volumeCommands.get_snapshots(),
        'create_snapshot':volumeCommands.create_snapshot(),
        'get_snapshot':volumeCommands.get_snapshot(),
        'update_snapshot':volumeCommands.update_snapshot(),
        'delete_snapshot':volumeCommands.delete_snapshot(),
        'list_snapshot_files':volumeCommands.list_snapshot_files()
    }


class Files(baseCommand.BaseCommandMap):
    """
    @help: run operations on tapis files
    """
    command_map = {
        'ls':fileCommands.ls(),
        'cd':fileCommands.cd(),
        'pwd':fileCommands.pwd(),
        'showme':fileCommands.showme(),
        'cat':fileCommands.cat(),
        'mkdir':fileCommands.mkdir(),
        'mv':fileCommands.mv(),
        'cp':fileCommands.cp(),
        'rm':fileCommands.rm(),
        #'get_recent_transfers':fileCommands.get_recent_transfers(),
        'create_postit':fileCommands.create_postit(),
        'list_postits':fileCommands.list_postits(),
        'get_postit':fileCommands.get_postit(),
        'delete_postit':fileCommands.delete_postit(),
        'redeem_postit':fileCommands.redeem_postit(),
        'upload':fileCommands.upload(),
        'download':fileCommands.download(),
    }


class Apps(baseCommand.BaseCommandMap):
    """
    @help: Run operations on tapis apps
    """
    command_map = {
        'create_app':appCommands.create_app(),
        'update_app':appCommands.update_app(),
        'get_apps':appCommands.get_apps(),
        'delete_app':appCommands.delete_app(),
        'get_app':appCommands.get_app(),
        'assign_default_job_attributes':appCommands.assign_default_job_attributes(),
        'is_app_enabled':appCommands.is_app_enabled(),
        'enable_app':appCommands.enable_app(),
        'disable_app':appCommands.disable_app(),
        'undelete_app':appCommands.undelete_app(),
        'get_app_history':appCommands.get_app_history(),
        'get_app_user_perms':appCommands.get_app_user_perms(),
        'grant_app_user_perms':appCommands.grant_app_user_perms(),
        'revoke_app_user_perms':appCommands.revoke_app_user_perms()
    }


class Jobs(baseCommand.BaseCommandMap):
    """
    @help: run jobs on a tapis system
    """
    command_map = {
        'hide_job':jobCommands.hide_job(),
        'unhide_job':jobCommands.unhide_job(),
        'cancel_job':jobCommands.cancel_job(),
        'get_job':jobCommands.get_job(),
        'get_job_history':jobCommands.get_job_history(),
        'get_jobs':jobCommands.get_jobs(),
        'download_job_output':jobCommands.download_job_output(),
        'get_job_status':jobCommands.get_job_status(),
        'resubmit_job':jobCommands.resubmit_job(),
        'submit_job':jobCommands.submit_job(),
        'share_job':jobCommands.share_job(),
        'get_job_share':jobCommands.get_job_share(),
        'delete_job_share':jobCommands.delete_job_share(),
        'subscribe_to_job':jobCommands.subscribe_to_job(),
        'get_subscriptions':jobCommands.get_subscriptions(),
        'delete_subscriptions':jobCommands.delete_subscriptions()
    }


class Query(baseCommand.BaseCommandMap):
    """
    @help: run integrated query CLIs
    """
    command_map = {
        'postgres': postgres.postgres(),
        'neo4j': neo4j.neo4j(),
        'get_neo4j_pod_uri':neo4j.get_neo4j_pod_uri(),
    }


class ArgsGenerator:
    def generate_truncated_arguments(self, arg_dict: dict) -> dict:
        truncation_dict = dict()
        for argument_name in arg_dict.keys():
            truncated_argument = self.__generate_truncated_argument(argument_name, truncation_dict)
            truncation_dict[argument_name] = truncated_argument
        return truncation_dict
            
    def __generate_truncated_argument(self, argument, truncated_arguments_dict, attempts=1):
        if argument[attempts-1] in ('_', '-'):
            argument += 1
        abbreviation = argument[:attempts]
        for index, letter in enumerate(argument):
            if letter.isupper() or not (argument[index-1].isnumeric() or argument[index-1].isalpha()):
                abbreviation += letter.lower()

        if abbreviation.lower() in list(truncated_arguments_dict.values()):
            abbreviation = self.__generate_truncated_argument(argument, truncated_arguments_dict, attempts=attempts+1)
        return abbreviation.lower()
    

class AggregateCommandMap(baseCommand.CommandContainer, ArgsGenerator):
    groups = {
        'Systems': Systems(),
        'General': General(),
        'Pods': Pods(),
        'Files': Files(),
        'Apps': Apps(),
        'Query': Query(),
        'Volumes':Volumes(),
        'Jobs': Jobs()
    }
    def __init__(self):
        truncated_arguments = self.generate_truncated_arguments(self.arguments)
        for command in self.aggregate_command_map.values():
            command.update_args_with_truncated(truncated_arguments)
        for name, argument in self.arguments.items():
            argument.truncated_arg = truncated_arguments[name]
        self.help = self.__general_help()

    def __general_help(self):
        general_help = list()
        for group in self.groups.values():
            general_help.append(group.brief_help)
        return general_help

    def update_credentials(self, t, username, password):
        for command_name, command in self.aggregate_command_map.items():
            command.set_t_and_creds(t, username, password, self)

    async def run_command(self, connection, command_data: dict):
        """
        process and run command based on received kwargs
        """
        command_name = command_data['command_selection']

        if command_name in self.aggregate_command_map:
            command = self.aggregate_command_map[command_name]
            command_data['connection'] = connection
            kwargs = dict()
            for argument, value in command_data.items():
                if argument in command.arguments:
                    kwargs[argument] = value
            return await command(**kwargs)
        elif command_name in self.groups:
            return self.groups[command_name]()
        elif command_name == "help":
            return self.help
        else:
            raise exceptions.CommandNotFoundError(command_name)


import re
import inspect
import abc
import ast
import json
from typing import Type
import typing
from abc import abstractmethod, ABC
import os
from pprint import pprint
from datetime import datetime

from commands import decorators # I finally understand. Imported at the top level by serverRun, so it can only see packages from that vantage point
from utilities import exceptions
from commands.arguments.argument import Argument, DynamicChoiceList, ALLOWED_ARG_TYPES
from socketopts import schemas
from commands import dataFormatters


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(f"{__file__}")))
saved_command_root_dir = os.path.join(__location__, r'..\user_files')


def get_kwargs(function):
    sig = inspect.signature(function)
    keyword_args = [param.name for param in sig.parameters.values() if param.default != inspect.Parameter.empty]
    return keyword_args

def get_args(function):
    sig = inspect.signature(function)
    positional_args = [param.name for param in sig.parameters.values() if param.default == inspect.Parameter.empty]
    return positional_args


class HelpStringRetriever:
    def help_string_retriever(self):
        try:
            docstring_components = self.__class__.__doc__
            docstring_components = docstring_components.split("@")
            for docstring_component in docstring_components:
                if re.match(r'^[^:]+', docstring_component).group(0) == "help":
                    return docstring_component.split("help: ")[1]
            raise Exception()
        except Exception as e:
            raise exceptions.HelpDoesNotExist(self.__class__.__name__)
    

class CommandMetaClass(abc.ABCMeta):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('BaseQuery', 'BaseCommand'):
            instance.__check_run(name, attrs)
            instance.__check_command_args(name, attrs)
            instance.__check_decorator(name, attrs)
            instance.__check_command_opt(name, attrs)
        return instance
    
    def __check_run(self, name, attrs):
        if 'run' not in list(attrs.keys()):
            raise AttributeError(f"The command {name} requires a 'run()' method")
        elif not inspect.iscoroutinefunction(attrs['run']):
            raise AttributeError(f"The run method of the command {name} must be a coroutine")
    
    def __check_command_args(self, name, attrs):
        run_params = get_args(attrs['run'])
        args = []
        if 'required_arguments' in list(attrs.keys()): 
            args += attrs['required_arguments']
        if 'optional_arguments' in list(attrs.keys()): 
            args += attrs['optional_arguments']
        if args:
            for argument in args:
                if not issubclass(argument.__class__, Argument) and not isinstance(argument, Argument):
                    raise AttributeError(f"The argument {argument.argument} of the command '{name}' must be of type 'Argument'")
                if 'optional_arguments' in list(attrs.keys()) and argument in attrs['optional_arguments'] and argument.positional:
                    raise AttributeError(f"The optional argument {argument.argument} of the command {name} cannot be positional. All positional arguments must be required")
        if 'kwargs' not in run_params or 'args' not in run_params:
            raise AttributeError(f"The run() method of the {name} class must have a **kwargs and *args as parameters")
        
    def __check_decorator(self, name, attrs):
        if 'decorator' in list(attrs.keys()) and type(attrs['decorator']) not in decorators.DECORATOR_LIST:
            raise TypeError(f"The decorator parameter of the command {name} is invalid. Must be set to None or to a decorator type. Currently is {type(attrs['decorator'])}")
        
    def __check_command_opt(self, name, attrs):
        if 'command_opt' in list(attrs.keys()) and type((attrs['command_opt'])) != list:
            raise TypeError(f"The command opt attribute of the command {name} must be a list!")
        

class UpdatableFormRetriever(abc.ABC):
    @abc.abstractmethod
    def __call__(self, tapis_instance, **kwargs):
        pass


class HelpMenu:
    def __init__(self, required_arguments: dict[str, Argument], optional_arguments: dict[str, Argument]):
        self.required_arguments = required_arguments
        self.optional_arguments = optional_arguments
        self.arguments = {**required_arguments, **optional_arguments}
        self.help = self.create_help_menu()

    def create_help_menu(self):
        help_dict = {'required':dict(), 'optional':dict()}
        for arg_name, argument in self.required_arguments.items():
            help_dict['required'][arg_name] = argument.help_message()
        for arg_name, argument in self.optional_arguments.items():
            help_dict['optional'][arg_name] = argument.help_message()
        return help_dict
    
    def dict(self):
        return self.arguments


class BaseCommand(ABC, HelpStringRetriever, metaclass=CommandMetaClass):
    decorator = None
    return_fields: list = []
    command_opt: list = None
    supports_config_file: bool = False
    required_arguments: list[Argument] | dict = list()
    optional_arguments: list[Argument] | dict = list()
    updateable_form_retriever: UpdatableFormRetriever = None
    def __init__(self):
        self.t = None
        self.username = None
        self.password = None
        self.server = None
        self.arguments = dict()
        self.return_formatter: dataFormatters.BaseDataFormatter = dataFormatters.BaseDataFormatter(self.return_fields)
        self.default_arguments()
        if self.supports_config_file:
            self.optional_arguments.append(Argument('file'))
        if isinstance(self.required_arguments, list):
            self.required_arguments = {argument.argument:argument for argument in self.required_arguments}
            self.arguments.update(**self.required_arguments)
        if isinstance(self.optional_arguments, list):  
            self.optional_arguments = {argument.argument:argument for argument in self.optional_arguments}
            self.arguments.update(**self.optional_arguments)
        self.positional_arguments = [arg_name for arg_name, arg in self.required_arguments.items() if arg.positional]
        self.form_arguments = [arg_name for arg_name, arg in self.arguments.items() if arg.arg_type not in  ('standard', 'silent')]
        self.help: dict[dict[str, list[dict[str, str]]]] = dict()

        self.command_execution_sequence = []
        if self.positional_arguments:
            self.command_execution_sequence.append(self.check_for_positionals)
        if self.command_opt:
            self.command_execution_sequence.append(self.handle_arg_opts)
        self.command_execution_sequence.append(self.verify_argument_rules_followed)
        if self.form_arguments:
            self.command_execution_sequence.append(self.handle_form_input)
        self.command_execution_sequence.append(self.verify_argument_rules_followed)
        self.command_execution_sequence.append(self.filter_kwargs)

    def default_arguments(self):
        """
        these are pseudarguments that allow some non-argument data to be passed to the command. This is a bit hacky, but its programatically necessary in the context of this framework
        """
        default_arguments = [Argument('connection', arg_type='silent'),
                            Argument('verbose', arg_type='silent'),
                            Argument('help', arg_type='silent'),
                            Argument('positionals', arg_type='silent')]
        
        for command in default_arguments:
            self.required_arguments.append(command)

    async def verify_argument_rules_followed(self, kwargs):
        """
        verify that the rules for each argument are followed, this is the argument validation
        """
        for name, value in self.arguments.items():
            if name in self.required_arguments and kwargs[name] == None:
                raise Exception(f"The argument {name} is required by the command {self.__class__.__name__}")
            elif name in kwargs and name in self.required_arguments and name not in self.form_arguments:
                kwargs[name] = self.required_arguments[name].verify_standard_value(kwargs[name])
            elif name in kwargs and name in self.optional_arguments and name not in self.form_arguments:
                kwargs[name] = self.optional_arguments[name].verify_standard_value(kwargs[name])
            if kwargs[name] and value.depends_on:
                for dependency in value.depends_on:
                    if not kwargs[dependency]:
                        raise Exception(f"The argument {name} requires the arguments {value.depends_on}")
        return kwargs

    async def filter_kwargs(self, kwargs):
        """
        filters out kwargs that have None value, Tapis breaks if I dont do this
        """
        filtered_kwargs = dict()
        for arg, value in kwargs.items():
            if arg in self.arguments and value or arg in self.required_arguments or (value == False and arg in self.arguments and self.arguments[arg].arg_type == 'standard'):
                filtered_kwargs[arg] = value
        return filtered_kwargs

    async def handle_config_file(self, kwargs):
        """
        when there is a config file submitted, that config file overrides all the other received kwargs
        """
        if kwargs['file']:
            with open(kwargs['file'], 'r') as f:
                kwargs = json.load(f)
            if 'error' in kwargs:
                kwargs.pop('error')
        return kwargs
    
    async def handle_form_input(self, kwargs, complex_args_flag=True):
        """
        handles forms when the user selects to fill them out
        """
        existing_values = dict()
        if self.updateable_form_retriever:
            existing_values = self.return_formatter.obj_to_dict(self.updateable_form_retriever(self.t, **kwargs))
        for arg_name in self.form_arguments:
            if kwargs[arg_name] or arg_name in self.required_arguments:
                request = schemas.FormRequest(request_content={arg_name:self.arguments[arg_name]}, existing_data=existing_values)
                await kwargs['connection'].send(request)
                response: schemas.FormResponse = await kwargs['connection'].receive()
                kwargs.update(**response.request_content)
        return kwargs

    async def handle_arg_opts(self, kwargs):
        """
        arg opts handle special operations, like supporting relative file paths, and system entry
        """
        for opt in self.command_opt:
            kwargs = opt(kwargs)
        return kwargs
    
    async def check_for_positionals(self, kwargs):
        """
        some commands contain positional arguments. This function handles that circumstance
        """
        for arg_name, value in zip(self.positional_arguments, kwargs['positionals']):
            kwargs[arg_name] = value
        return kwargs

    def set_t_and_creds(self, t, username, password, server):
        """
        whenever the tenant or user changes, the new tapis object with the new credentials must be passed to each command to ensure they are operating on the currect user
        """
        self.t = t
        self.username = username
        self.password = password
        self.server = server
        for key, arg in self.optional_arguments.items():
            if issubclass(arg.choices.__class__, DynamicChoiceList):
                arg.choices = arg.choices(self.t)

    def update_args_with_truncated(self, truncated_args_dict):
        """
        when the command group finishes processing all the truncated arguments, they get passed back here to be processed and assigned
        """
        try:
            for key in self.required_arguments:
                self.required_arguments[key].truncated_arg = f"-{truncated_args_dict[key]}"
            for key in self.optional_arguments:
                self.optional_arguments[key].truncated_arg = f"-{truncated_args_dict[key]}"
            self.help['required'] = self.__argument_help_compiler(self.required_arguments)
            self.help['optional'] = self.__argument_help_compiler(self.optional_arguments)
        except Exception as e:
            print(self.__class__.__name__)
            raise e
        
    def __argument_help_compiler(self, arg_dict: dict[str, Argument]):
        """
        compile the help menus based on stored command metadata
        """
        verbose_dict = dict()
        standard_str = str()
        for name, argument in arg_dict.items():
            if not (name in self.required_arguments and name in self.form_arguments) and argument.arg_type != 'silent':
                verbose_dict[name] = argument.help_message()
                standard_str += argument.str()
        return {'verbose':verbose_dict, 'standard':standard_str}

    def __get_help(self, verbose):
        """
        return the compiled help menus
        """
        help_dict = {"Command":self.__class__.__name__,
                    "Description":self.help_string_retriever()}
        if not verbose:
            help_dict.update(**{
                "Syntax":f"{self.__class__.__name__} {self.help['required']['standard']}\n(Optional Arguments) {self.help['optional']['standard']}"})
        else:
            help_dict.update(**{
                "Syntax":f"{self.__class__.__name__}",
                "Required Arguments":self.help['required']['verbose'],
                "Optional Arguments":self.help['optional']['verbose']})
        return help_dict
    
    def brief_help(self):
        """
        return non verbose help
        """
        return {"Command":self.__class__.__name__,
                "Description":self.help_string_retriever()}
    
    @abstractmethod
    async def run(self, *args, **kwargs):
        """
        contains the actual command information at command definition
        """
        pass

    async def __call__(self, **kwargs):
        """
        runs all command meta-operations
        """
        if self.supports_config_file and kwargs['file']:
            new_kwargs = await self.handle_config_file(kwargs)
            for arg_name, arg in kwargs.items():
                if arg_name in ('connection', 'positionals', 'verbose', 'help'):
                    new_kwargs[arg_name] = arg
            kwargs = new_kwargs
        else:
            if self.decorator:
                try:
                    return_value = await self.decorator(input_command=self, **kwargs)
                except (ValueError, exceptions.NoConfirmationError) as e:
                    return f"Command execution failed due to {e}"

            if kwargs['help']:
                return self.__get_help(verbose=kwargs['verbose'])
            for handler in self.command_execution_sequence:
                kwargs = await handler(kwargs)
        try:
            return_value = await self.run(**kwargs)
        except (exceptions.Shutdown, exceptions.Exit) as e:
            raise e
        except Exception as e:
            if self.supports_config_file:
                current_date = datetime.now()
                formatted_date = current_date.strftime("%m-%d-%y")
                main_save_path = f"{saved_command_root_dir}\\{formatted_date}"
                try:
                    os.listdir(main_save_path)
                except:
                    os.makedirs(main_save_path)
                number = len(os.listdir(main_save_path))
                file_save_path = f"{main_save_path}\\{self.__class__.__name__}_{number}.json"
                with open(file_save_path, 'w') as f:
                    kwargs.pop('connection')
                    kwargs.pop('positionals')
                    kwargs.pop('verbose')
                    kwargs.pop('help')
                    kwargs['error'] = str(e)
                    json.dump(kwargs, f, indent=4)
                    raise Exception(e, f"Argument input failure, command data written to file {file_save_path}")
            raise e
        return_value = self.return_formatter(return_value, kwargs['verbose'])
        return return_value
    

class BaseQuery(BaseCommand):
    def get_pod_credentials(self, id):
        uname, pword = self.t.pods.get_pod_credentials(pod_id=id).user_username, self.t.pods.get_pod_credentials(pod_id=id).user_password
        return uname, pword
    
    @abstractmethod
    async def run(self, *args, **kwargs):
        pass
    

class CommandMapMetaClass(type):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('CommandMapMetaClass', 'BaseCommandMap'):
            instance.__check_commands_are_proper_type(name, attrs)
            instance.__check_command_name(name, attrs)
            instance.__special_command_opts(name, attrs)
        return instance
    
    def __check_commands_are_proper_type(self, name, attrs):
        commands = attrs['command_map']
        if not commands:
            raise AttributeError(f"The command group {name} has no commands!")
            
    def __check_command_name(self, name, attrs):
        commands = attrs['command_map']
        for command_name, command in commands.items():
            if command_name != command.__class__.__name__:
                raise AttributeError(f"The command {command_name} in the command map {name} is a different name from its corresponding command class, {command.__class__.__name__}")

    def __special_command_opts(self, name, attrs):
        if 'command_opt' in list(attrs.keys()):
            if type(attrs['command_opt']) != list:
                raise TypeError(f"The command_opt attribute for the command group {name} must be a list!")
            for command_name, command in attrs['command_map'].items():
                if not command.command_opt:
                    command.command_opt = attrs['command_opt']
                    continue
                command.command_opt += attrs['command_opt']


class CommandContainer:
    aggregate_command_map: dict[str, Type[BaseCommand]] = dict()
    arguments : dict[str, Argument]= {
        'help':Argument('help', action='store_true'),
        'verbose':Argument('verbose', action='store_true'),
        'file':Argument('file')
    }


class BaseCommandMap(CommandContainer, HelpStringRetriever, metaclass=CommandMapMetaClass):
    command_map: dict[str, Type[BaseCommand]] = None
    command_opt = None
    def __init__(self):
        self.brief_help = self.__brief_help_gen()
        self.verbose_help = self.__help_gen()
        for name, command in self.command_map.items():
            self.aggregate_command_map.update({name:command})
            if command.required_arguments or command.optional_arguments:
                self.__contribute_to_arguments(command)

    def __contribute_to_arguments(self, command):
        arg_dict = {**command.required_arguments, **command.optional_arguments}
        for name, arg in arg_dict.items():
            if name not in self.arguments:
                self.arguments[name] = arg
            elif self.arguments[name].check_for_copy_data() != arg.check_for_copy_data():
                raise ValueError(f"Found two instances of the argument {arg.argument} that had different attributes.\n\n{command.__class__.__name__}: {arg.check_for_copy_data()}\n\nvs\n\n{self.arguments[name].check_for_copy_data()}")

    def __help_gen(self):
        verbose_help = list()
        for command in self.command_map.values():
            verbose_help.append(command.brief_help())
        return verbose_help
    
    def __brief_help_gen(self):
        brief_help = {
            'group':self.__class__.__name__,
            'description':self.help_string_retriever(),
            'instruction':f"enter {self.__class__.__name__} to retrieve list of commands"
        }
        return brief_help
    
    def __call__(self):
        return self.verbose_help



"""
DECORATORS
These decorators are used in the tapisObjectWrappers.py file to standardize special functions. Allows for increased code reusability
"""
import typing
import time
import abc
import socket
from functools import update_wrapper, partial

try:
    from socketopts import schemas
    from utilities import exceptions, killableThread
    from commands.arguments import argument
except:
    from ..socketopts import schemas
    from ..utilities import exceptions, killableThread
    from ..commands.arguments import argument


class BaseRequirementDecorator(abc.ABC):
    username: typing.Optional[str] = None
    password: typing.Optional[str] = None

    @abc.abstractmethod
    async def __call__(self, input_ommand: typing.Type[object], connection: socket.socket, *args, **kwargs):
        pass


class Auth(BaseRequirementDecorator):
    """
    used for secure authentication from the client. Requires that the function has a username and password parameter for credentials. sends request for credentials from 
    the client, and checks those credentials against the stored credentials in the server.
    """
    async def __call__(self, input_command, *args, **kwargs):
        connection = kwargs['connection']
        requires_username = True
        if connection:
            auth_request = schemas.FormRequest(request_content=argument.Form('auth', [
                        argument.Argument('username'),
                        argument.Argument('password', arg_type='secure')
                    ]).json(),
                    message={"message":"Password is required to continue. If you logged in using TACC password login, use that. Otherwise use the session password\nYour username was returned during initial login"})
            await connection.send(auth_request)
            auth_data = await connection.receive()
            if auth_data.request_content['password'] != self.password:
                raise ValueError("The provided password does not match the stored password")
            if auth_data.request_content['username'] != self.username:
                raise ValueError("The provided username does not match the stored username")
            return await input_command.run(**kwargs)
        return await input_command.run(**kwargs)


class NeedsConfirmation(BaseRequirementDecorator):
    """
    add to functions that you want user confirmation to exit. If you accidentally enter a command to delete a pod, this will not let you until you confirm
    """
    async def __call__(self, input_command, *args, **kwargs):
        connection = kwargs['connection']
        if connection:
            connection = connection
            confirmation_request = schemas.FormRequest(message={"message":f"YOU REQUESTED TO {input_command.__class__.__name__.upper()}. THIS MIGHT CAUSE DATA LOSS! Please confirm (y/n)"},
                                                       request_content={"confirmation":argument.Argument('confirm', arg_type='confirmation')})
            await connection.send(confirmation_request)
            confirmation_reply: schemas.FormResponse = await connection.receive()
            confirmed = confirmation_reply.request_content['confirmation']
            if not confirmed:
                raise exceptions.NoConfirmationError(input_command)
        return await input_command.run(**kwargs)


DECORATOR_LIST = [
    NeedsConfirmation,
    Auth
]
    
class DecoratorSetup:
    """
    for instantiation of the tapis wrappers, and the server, to set up decorators with user credentials and the socket connection. If you want to use the decorators in your class
    YOU WILL NEED TO USE THIS!
    """
    def configure_decorators(self, username, password):
        self.username = username
        self.password = password


class AnimatedLoading:
    """
    Add this if you want to print a loading animation while a function is executing
    """
    def __init__(self, func: typing.Callable):
        update_wrapper(self, func)
        self.func = func
        self.__code__ = func.__code__
        self.animation_frames = ['|','/','-','\\']

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)

    def __repr__(self):
        return self.func
    
    def __str__(self):
        return str(self.func)
    
    def animation(self):
        while True:
            for frame in self.animation_frames:
                print(f"loading {frame}", end='', flush=True)
                time.sleep(0.5)
    
    def __call__(self, obj, *args, **kwargs):
        animation_thread = killableThread.KillableThread(target=self.animation)
        animation_thread.start()
        result = self.func(obj, *args, **kwargs)
        animation_thread.kill()
        print("", end='', flush=True)
        return result


import logging
import sys
import typing


class ServerLogger:
    def initialize_logger(self, name):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.INFO)
        stream_handler = logging.StreamHandler(stream=sys.stdout)

        file_handler = logging.FileHandler(
            r'logs.log', mode='w')
        stream_handler.setLevel(logging.INFO)
        file_handler.setLevel(logging.INFO)

        # set formats
        stream_format = logging.Formatter(
            '%(name)s - %(levelname)s - %(message)s')
        file_format = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        stream_handler.setFormatter(stream_format)
        file_handler.setFormatter(file_format)

        # add the handlers
        self.logger.addHandler(stream_handler)
        self.logger.addHandler(file_handler)

        self.logger.disabled = False


class ConnectionLogger:
    def initialize_logger(self, name):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.INFO)
        stream_handler = logging.StreamHandler(stream=sys.stdout)
        stream_handler.setLevel(logging.INFO)
        stream_format = logging.Formatter(
            '%(name)s - %(levelname)s - %(message)s')
        stream_handler.setFormatter(stream_format)
        self.logger.addHandler(stream_handler)
        self.logger.disabled = False


class ServiceLogger:
    levels_map = {
        'info':logging.INFO,
        'warning':logging.WARNING,
        'error':logging.ERROR
    }
    def __init__(self, name, log_path):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.INFO)
        self.file_handler = logging.FileHandler(
            log_path, mode='a')
        self.file_handler.setLevel(logging.INFO)
        file_format = logging.Formatter(
            '%(name)s - %(levelname)s - %(message)s')
        self.file_handler.setFormatter(file_format)
        self.logger.addHandler(self.file_handler)
        self.logger.disabled = False

    def log(self, data, level: typing.Literal['info', 'warning', 'error'], filename: str):
        self.file_handler.baseFilename = filename
        level = self.levels_map[level]
        self.logger.log(level, data)


import time
import socket
import asyncio
import traceback
import typing
import os
import json

from tapipy.tapis import Tapis

from commands import commandMap, decorators
from utilities import logger, exceptions
from socketopts import schemas, socketOpts
from server import auth


class ServerConnection(socketOpts.ServerSocketOpts):
    """
    connection object to wrap around async reader and writer to make work easier
    """
    def __init__(self, name, reader, writer, connection_list, debug=False):
        super().__init__(name, debug=debug)
        self.name = name
        self.connection_list: list = connection_list
        self.reader: asyncio.StreamReader = reader
        self.writer: asyncio.StreamWriter = writer
        self.task: asyncio.Task = None
        self.status = "CLOSED"
        self.shutdown_message = schemas.ResponseData(message={'message':'Shutdown initiated, closing'}, exit_status=1)

    def set_status_device_authenticating(self):
        self.status = 'DEVICE_CODE_AUTH'

    def set_status_closed(self):
        self.status = 'CLOSED'

    def set_status_exiting(self):
        self.status = 'EXITING'

    def set_task(self, task: asyncio.Task):
        self.task = task
        self.status = 'OPEN'

    async def close(self, connection_list_lock: asyncio.Lock):
        self.logger.info('ATTEMPTING TO CLOSE')
        if self.status in ('OPEN', 'EXITING'):
            result = self.task.cancel()
            self.logger.info("successfully cancelled task")
            await self.send(self.shutdown_message)
            self.writer.close()
            await self.writer.wait_closed()
            self.status = 'CLOSED'
        elif self.status == 'DEVICE_CODE_AUTH': # only in the auth file
            await self.send(schemas.AuthRequest(error='cancelled authentication during device_code grant', auth_request_type='device_code'))
            self.writer.close()
            await self.writer.wait_closed()
        else:
            self.writer.close()
            await self.writer.wait_closed()
        async with connection_list_lock:
            self.connection_list.remove(self)
        print(self.connection_list)
        self.logger.info('SUCCESSFULLY CLOSED')


class Server(commandMap.AggregateCommandMap, logger.ServerLogger, decorators.DecoratorSetup, auth.ServerSideAuth):
    """
    Receives commands from the client and executes Tapis operations
    """
    SESSION_TIME = 5000
    debug=False
    def __init__(self, IP: str, PORT: int):
        super().__init__()
        self.initial = True
        self.running = True

        self.t = None
        self.url = None
        self.access_token = None
        self.username = None
        self.password = None
        self.auth_type = None

        self.__name__ = "Server"
        self.initialize_logger(self.__name__)
        # setting up socket server
        self.ip, self.port = IP, PORT
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setblocking(False)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.ip, self.port))
        self.sock.listen(1)

        self.end_time = time.time() + self.SESSION_TIME # start the countdown on the timeout

        self.connections_list: list[ServerConnection] = []

        self.server = None
        self.num_connections = 0

        self.connection_list_lock = asyncio.Lock()
        self.logger.info('initialization complete')

    async def close_connections(self):
        print([connection.status for connection in self.connections_list])
        return await asyncio.gather(*[connection.close(self.connection_list_lock) for connection in self.connections_list])

    async def close(self):
        await self.close_connections()
        self.server.close()
        raise exceptions.Shutdown()

    async def check_timeout(self):
        while self.running:
            if time.time() >= self.end_time:
                self.logger.info("Timeout, shutting down")
                self.running = False
                return 'server timed out'
            await asyncio.sleep(3)
    
    async def check_shutdown(self):
        while self.running:
            await asyncio.sleep(3)
            for connection in self.connections_list:
                if connection.status == 'EXITING':
                    await connection.close(self.connection_list_lock)
        await self.close()
        self.logger.info("The server shutdown.")
        return None

    async def handshake(self, connection):
        self.logger.info("Handshake starting")
        if self.initial:  # if this is the first time in the session that the cli is connecting
            startup_data = schemas.StartupData(initial = self.initial)
            await connection.send(startup_data)
            await self.auth_startup(connection)
        else:
            startup_result = schemas.StartupData(initial = self.initial, username = self.username, url = self.url)
            await connection.send(startup_result)
        self.initial = False
        self.logger.info("Final connection data sent")

    async def accept(self, reader, writer):
        """
        accept connection request and initialize communication with the client
        """  
        self.num_connections += 1
        self.end_time = time.time() + self.SESSION_TIME
        connection = ServerConnection(f"CON-{self.num_connections}", reader=reader, writer=writer, connection_list=self.connections_list, debug=self.debug)
        ip, port = writer.transport.get_extra_info('socket').getsockname()
        self.logger.info("Received connection request")
        try:
            await self.handshake(connection)
        except exceptions.InvalidCredentialsReceived as e:
            self.logger.warning("invalid credentials entered too many times. Cancelling request")
            await connection.close(self.connection_list_lock)
            return
        except exceptions.ClientSideError as e:
            self.logger.warning(f"Encountered client side error during startup handshake. {e}")
            await connection.close(self.connection_list_lock)
            return
        except Exception as e:
            self.logger.warning(e)
            await connection.close(self.connection_list_lock)
            return
        self.logger.info("connection is running now")
        await connection.send(schemas.ResponseData(request_content={name:argument.json() for name, argument in self.arguments.items()}))
        
        setup_response: schemas.ResponseData = await connection.receive()
        if not setup_response.request_content['setup_success']:
            self.logger.warning(f"The setup of the connection {connection.name} failed")
            return

        loop = asyncio.get_event_loop()
        task: asyncio.Task = loop.create_task(self.receive_and_execute(connection))
        connection.set_task(task)
        self.connections_list.append(connection)
        print([connection.status for connection in self.connections_list])

    async def receive_and_execute(self, connection: ServerConnection):
        """
        receive and process commands
        """
        self.logger.info(f"{connection.name} is now running")
        while self.running:
            try:
                message = await connection.receive()
                if not self.running:
                    raise exceptions.Shutdown
                kwargs = message.request_content
                result = await self.run_command(connection, kwargs)
                response = schemas.ResponseData(message={"message":result}, url=self.url, active_username=self.username)
                self.end_time = time.time() + self.SESSION_TIME 
                await connection.send(response)
                self.logger.info(message.schema_type)
            except exceptions.ClientSideError as e:
                self.logger.warning(e)
                continue
            except (exceptions.TimeoutError, exceptions.Shutdown) as e:
                #error_response = schemas.ResponseData(error=str(e), url=self.url, active_username=self.username, exit_status=1)
                #await connection.send(error_response)
                self.logger.warning(str(e))
                self.running = False
                return
            except exceptions.Exit as e:
                self.logger.info("user exit initiated")
                # error_response = schemas.ResponseData(error=str(e), exit_status=1, url=self.url, active_username=self.username)
                # await connection.send(error_response)
                connection.set_status_exiting()
                return
            except OSError:
                self.logger.info("connection was lost, waiting to reconnect")
                await connection.close(self.connection_list_lock)
            except asyncio.CancelledError:
                return 'task was cancelled'
            except (exceptions.CommandNotFoundError, exceptions.NoConfirmationError, exceptions.InvalidCredentialsReceived, Exception) as e:
                error_str = traceback.format_exc()
                error_response = schemas.ResponseData(error=str(e), url=self.url, active_username=self.username)
                await connection.send(error_response)
                self.logger.warning(f"{error_str}")
    
    async def main(self):
        self.server = await asyncio.start_server(self.accept, sock=self.sock)
        try:
            async with self.server:
                result = await asyncio.gather(self.server.serve_forever(), self.check_timeout(), self.check_shutdown(), return_exceptions=True)#, self.check_timeout(), return_exceptions=True)
                self.logger.info(str(result))
        except (KeyboardInterrupt, exceptions.Shutdown):
            self.running = False


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    server = Server(socket.gethostbyname('127.0.0.1', 30000)) 
    try:
        asyncio.run(server.main())
    finally:
        loop.close()
